=== modified file 'lib/canonical/codehosting/tests/helpers.py'
--- lib/canonical/codehosting/tests/helpers.py	2008-04-15 23:25:31 +0000
+++ lib/canonical/codehosting/tests/helpers.py	2008-04-16 13:16:57 +0000
@@ -7,10 +7,11 @@
     'AvatarTestCase', 'CodeHostingTestProviderAdapter',
     'CodeHostingRepositoryTestProviderAdapter', 'FakeLaunchpad',
     'ServerTestCase', 'adapt_suite', 'create_branch_with_one_revision',
-    'make_bazaar_branch_and_tree']
+    'deferToThread', 'make_bazaar_branch_and_tree']
 
 import os
 import shutil
+import threading
 import unittest
 
 import transaction
@@ -32,7 +33,8 @@
 from canonical.launchpad.webapp.authorization import LaunchpadSecurityPolicy
 from canonical.testing import LaunchpadFunctionalLayer, TwistedLayer
 
-from twisted.internet import defer
+from twisted.internet import defer, threads
+from twisted.python.util import mergeFunctionMetadata
 from twisted.trial.unittest import TestCase as TrialTestCase
 from twisted.web.xmlrpc import Fault
 
@@ -240,6 +242,21 @@
         return self.server.getTransport(relpath)
 
 
+def deferToThread(f):
+    """Run the given callable in a separate thread and return a Deferred which
+    fires when the function completes.
+    """
+    def decorated(*args, **kwargs):
+        d = defer.Deferred()
+        def runInThread():
+            return threads._putResultInDeferred(d, f, args, kwargs)
+
+        t = threading.Thread(target=runInThread)
+        t.start()
+        return d
+    return mergeFunctionMetadata(f, decorated)
+
+
 class FakeLaunchpad:
     """Stub RPC interface to Launchpad.
 

=== modified file 'lib/canonical/launchpad/components/externalbugtracker/trac.py'
--- lib/canonical/launchpad/components/externalbugtracker/trac.py	2008-04-16 13:01:18 +0000
+++ lib/canonical/launchpad/components/externalbugtracker/trac.py	2008-04-16 19:49:35 +0000
@@ -7,6 +7,7 @@
 
 import csv
 from datetime import datetime
+import time
 import urllib2
 import xmlrpclib
 
@@ -271,6 +272,26 @@
         trac_time = datetime.fromtimestamp(utc_time)
         return trac_time.replace(tzinfo=pytz.timezone('UTC'))
 
+    @needs_authentication
+    def getModifiedRemoteBugs(self, remote_bug_ids, last_checked):
+        """See `IExternalBugTracker`."""
+        endpoint = urlappend(self.baseurl, 'xmlrpc')
+        server = xmlrpclib.ServerProxy(
+            endpoint, transport=self.xmlrpc_transport)
+
+        # Convert last_checked into an integer timestamp (which is what
+        # the Trac LP plugin expects).
+        last_checked_timestamp = int(
+            time.mktime(last_checked.timetuple()))
+
+        # We retrieve only the IDs of the modified bugs from the server.
+        criteria = {
+            'modified_since': last_checked_timestamp,
+            'bugs': remote_bug_ids,}
+        time_snapshot, modified_bugs = server.launchpad.bug_info(
+            0, criteria)
+
+        return [bug['id'] for bug in modified_bugs]
 
 class TracXMLRPCTransport(xmlrpclib.Transport):
     """XML-RPC Transport for Trac bug trackers.

=== modified file 'lib/canonical/launchpad/doc/externalbugtracker-trac-lp-plugin.txt'
--- lib/canonical/launchpad/doc/externalbugtracker-trac-lp-plugin.txt	2008-04-15 14:42:03 +0000
+++ lib/canonical/launchpad/doc/externalbugtracker-trac-lp-plugin.txt	2008-04-16 13:31:46 +0000
@@ -152,3 +152,72 @@
     Using XML-RPC to generate token.
     Successfully validated the token.
     datetime.datetime(2008, 4, 9, 2, 2, 1, tzinfo=<UTC>)
+
+
+== Getting modified bugs ==
+
+We only want to update the bug watches whose remote bugs have been
+modified since the last time we checked.
+
+In order to demonstrate this, we'll create some mock remote bugs for our
+test XML-RPC transport to check.
+
+    >>> from canonical.launchpad.ftests.externalbugtracker import (
+    ...     MockTracRemoteBug)
+
+    >>> remote_bugs = {
+    ...     '1': MockTracRemoteBug('1', datetime(2008, 4, 1, 0, 0, 0)),
+    ...     '2': MockTracRemoteBug('2', datetime(2007, 1, 1, 1, 1, 1)),
+    ...     '3': MockTracRemoteBug('3', datetime(2008, 1, 1, 1, 2, 3)),
+    ...     }
+
+    >>> test_transport.remote_bugs = remote_bugs
+
+Calling the getModifiedRemoteBugs() method of our Trac instance and
+passing it a list of bug IDs and a datetime object will return a list
+of the IDs of the bugs which have been modified since that time.
+
+    >>> bug_ids_to_check = ['1', '2', '3']
+    >>> last_checked = datetime(2008, 1, 1, 0, 0, 0)
+    >>> test_transport.expireCookie(test_transport.auth_cookie)
+    >>> sorted(trac.getModifiedRemoteBugs(
+    ...     bug_ids_to_check, last_checked))
+    Using XML-RPC to generate token.
+    Successfully validated the token.
+    ['1', '3']
+
+Different last_checked times will result in different numbers of bugs
+being returned.
+
+    >>> last_checked = datetime(2008, 2, 1, 0, 0, 0)
+    >>> test_transport.expireCookie(test_transport.auth_cookie)
+    >>> trac.getModifiedRemoteBugs(
+    ...     bug_ids_to_check, last_checked)
+    Using XML-RPC to generate token.
+    Successfully validated the token.
+    ['1']
+
+If no bugs have been updated since last_checked, getModifiedRemoteBugs()
+will return an empty list.
+
+    >>> last_checked = datetime(2008, 5, 1, 0, 0, 0)
+    >>> test_transport.expireCookie(test_transport.auth_cookie)
+    >>> trac.getModifiedRemoteBugs(
+    ...     bug_ids_to_check, last_checked)
+    Using XML-RPC to generate token.
+    Successfully validated the token.
+    []
+
+If we ask for bug ids that don't exist on the remote server, they will
+also be returned. This is so that when we try to retrieve the status of
+the missing bugs an error will be raised that we can then investigate.
+
+    >>> bug_ids_to_check = ['1', '2', '3', '99', '100']
+    >>> last_checked = datetime(2008, 1, 1, 0, 0, 0)
+    >>> test_transport.expireCookie(test_transport.auth_cookie)
+    >>> sorted(trac.getModifiedRemoteBugs(
+    ...     bug_ids_to_check, last_checked))
+    Using XML-RPC to generate token.
+    Successfully validated the token.
+    ['1', '100', '3', '99']
+

=== modified file 'lib/canonical/launchpad/ftests/externalbugtracker.py'
--- lib/canonical/launchpad/ftests/externalbugtracker.py	2008-04-16 13:01:18 +0000
+++ lib/canonical/launchpad/ftests/externalbugtracker.py	2008-04-16 19:49:35 +0000
@@ -11,6 +11,8 @@
 import urlparse
 import xmlrpclib
 
+from datetime import datetime
+
 from zope.component import getUtility
 
 from canonical.config import config
@@ -417,6 +419,14 @@
         return open(file_path + '/' + 'trac_example_ticket_export.csv', 'r')
 
 
+class MockTracRemoteBug:
+    """A mockup of a remote Trac bug."""
+
+    def __init__(self, id, last_modified):
+        self.id = id
+        self.last_modified = last_modified
+
+
 class TestTracInternalXMLRPCTransport:
     """Test XML-RPC Transport for the internal XML-RPC server.
 
@@ -442,6 +452,7 @@
 class TestTracXMLRPCTransport(TracXMLRPCTransport):
     """An XML-RPC transport to be used when testing Trac."""
 
+    remote_bugs = {}
     seconds_since_epoch = None
     local_timezone = 'UTC'
     utc_offset = 0
@@ -488,6 +499,82 @@
         utc_time = local_time - self.utc_offset
         return [self.local_timezone, local_time, utc_time]
 
+    def bug_info(self, level, criteria=None):
+        """Return info about a bug or set of bugs.
+
+        :param level: The level of detail to return about the bugs
+            requested. This can be one of:
+            0: Return IDs only.
+            1: Return Metadata only.
+            2: Return Metadata + comments.
+            3: Return all data about each bug.
+
+        :param criteria: The selection criteria by which bugs will be
+            returned. Possible keys include:
+            modified_since: An integer timestamp. If specified, only
+                bugs modified since this timestamp will
+                be returned.
+            bugs: A list of bug IDs. If specified, only bugs whose IDs are in
+                this list will be returned.
+
+        Return a list of [ts, bugs] where ts is a utc timestamp as
+        returned by `time_snapshot()` and bugs is a list of bug dicts.
+        """
+        # XXX 2008-04-12 gmb:
+        #     This is only a partial implementation of this; it will
+        #     grow over time as implement different methods that call
+        #     this method. See bugs 203564, 158703 and 158705.
+
+        # We sort the list of bugs for the sake of testing.
+        bug_ids = sorted([bug_id for bug_id in self.remote_bugs.keys()])
+        bugs_to_return = []
+        missing_bugs = []
+
+        for bug_id in bug_ids:
+            bugs_to_return.append(self.remote_bugs[bug_id])
+
+        if criteria is None:
+            criteria = {}
+
+        # If we have a modified_since timestamp, we return bugs modified
+        # since that time.
+        if 'modified_since' in criteria:
+            # modified_since is an integer timestamp, so we convert it
+            # to a datetime.
+            modified_since = datetime.fromtimestamp(
+                criteria['modified_since'])
+
+            bugs_to_return = [
+                bug for bug in bugs_to_return
+                if bug.last_modified > modified_since]
+
+        # If we have a list of bug IDs specified, we only return
+        # those members of bugs_to_return that are in that
+        # list.
+        if 'bugs' in criteria:
+            bugs_to_return = [
+                bug for bug in bugs_to_return
+                if bug.id in criteria['bugs']]
+
+            # We make a separate list of bugs that don't exist so that
+            # we can return them with a status of 'missing' later.
+            missing_bugs = [
+                bug_id for bug_id in criteria['bugs']
+                if bug_id not in self.remote_bugs]
+
+        # We only return what's required based on the level parameter.
+        # For level 0, only IDs are returned.
+        if level == 0:
+            bugs_to_return = [{'id': bug.id} for bug in bugs_to_return]
+
+        # Tack the missing bugs onto the end of our list of bugs. These
+        # will always be returned in the same way, no matter what the
+        # value of the level argument.
+        missing_bugs = [
+            {'id': bug_id, 'status': 'missing'} for bug_id in missing_bugs]
+
+        return [self.time_snapshot()[2], bugs_to_return + missing_bugs]
+
 
 class TestRoundup(Roundup):
     """Roundup ExternalBugTracker for testing purposes.

=== modified file 'lib/canonical/launchpad/ftests/trac-xmlrpc-transport.txt'
--- lib/canonical/launchpad/ftests/trac-xmlrpc-transport.txt	2008-04-11 10:05:15 +0000
+++ lib/canonical/launchpad/ftests/trac-xmlrpc-transport.txt	2008-04-16 11:48:27 +0000
@@ -54,3 +54,84 @@
 
     >>> server.launchpad.time_snapshot()
     ['US/Eastern', 1206328061, 1206342461]
+
+
+== launchpad.bug_info() ==
+
+bug_info() returns, as the name suggests, info about a given bug or set
+of bugs. It takes two parameters: level, an integer indicating how much
+data to return, and criteria, which specifies criteria by which to
+select the bugs to return.
+
+We'll add some bugs to our trac transport to demonstrate this.
+
+    >>> from datetime import datetime
+    >>> from canonical.launchpad.ftests.externalbugtracker import (
+    ...     MockTracRemoteBug)
+
+    >>> remote_bugs = {
+    ...     '1': MockTracRemoteBug(
+    ...         '1', datetime(2008, 4, 1, 0, 0, 0)),
+    ...     '2': MockTracRemoteBug(
+    ...         '2', datetime(2007, 1, 1, 1, 1, 1)),
+    ...     '3': MockTracRemoteBug(
+    ...         '3', datetime(2008, 1, 1, 1, 2, 3)),}
+
+    >>> trac_transport.remote_bugs = remote_bugs
+
+Specifying a level of 0 and no criteria will return the IDs of all the
+bugs, along with a time snapshot as returned by time_snapshot().
+
+    >>> time_snapshot, bugs = trac_transport.bug_info(level=0)
+    >>> print bugs
+    [{'id': '1'}, {'id': '2'}, {'id': '3'}]
+
+The criteria dict has two possible keys: modified_since and bugs.
+Specifying a value for modified_since will cause only the bugs modified
+since that time to be returned. modified_since is an integer timestamp,
+so we'll convert a datetime into one for the purposes of this test.
+
+    >>> import time
+    >>> last_checked = datetime(2008, 1, 1, 0, 0, 0)
+    >>> last_checked_timestamp = int(
+    ...     time.mktime(last_checked.timetuple()))
+
+    >>> criteria = {'modified_since': last_checked_timestamp}
+    >>> time_snapshot, bugs = trac_transport.bug_info(
+    ...     level=0, criteria=criteria)
+
+    >>> print bugs
+    [{'id': '1'}, {'id': '3'}]
+
+The bugs key in the criteria dict allows us to specify a list of bug IDs
+to return.
+
+    >>> criteria = {'bugs': ['1', '2']}
+    >>> time_snapshot, bugs = trac_transport.bug_info(
+    ...     level=0, criteria=criteria)
+
+    >>> print bugs
+    [{'id': '1'}, {'id': '2'}]
+
+If a bug doesn't exist, it will be returned with a status of
+'missing'.
+
+    >>> criteria = {'bugs': ['11', '12']}
+    >>> time_snapshot, bugs = trac_transport.bug_info(
+    ...     level=0, criteria=criteria)
+
+    >>> print bugs
+    [{'status': 'missing', 'id': '11'}, {'status': 'missing', 'id': '12'}]
+
+Combining the bugs and modified_since fields in the criteria dict will
+result in only the bugs modified since the modified_since time whose IDs
+are in the bugs list being returned.
+
+    >>> criteria = {
+    ...     'bugs': ['1', '2'],
+    ...     'modified_since': last_checked_timestamp,}
+    >>> time_snapshot, bugs = trac_transport.bug_info(
+    ...     level=0, criteria=criteria)
+
+    >>> print bugs
+    [{'id': '1'}]

